#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

# TODO: Disabled until we can confirm application to Pyomo models
import pyomo.solvers.plugins.solvers.CBCplugin
import pyomo.solvers.plugins.solvers.GLPK
import pyomo.solvers.plugins.solvers.CPLEX
import pyomo.solvers.plugins.solvers.GUROBI
import pyomo.solvers.plugins.solvers.BARON
import pyomo.solvers.plugins.solvers.ASL
import pyomo.solvers.plugins.solvers.pywrapper
import pyomo.solvers.plugins.solvers.SCIPAMPL
import pyomo.solvers.plugins.solvers.CONOPT
import pyomo.solvers.plugins.solvers.XPRESS
import pyomo.solvers.plugins.solvers.IPOPT
import pyomo.solvers.plugins.solvers.gurobi_direct
import pyomo.solvers.plugins.solvers.gurobi_persistent
import pyomo.solvers.plugins.solvers.cplex_direct
import pyomo.solvers.plugins.solvers.cplex_persistent
import pyomo.solvers.plugins.solvers.GAMS
import pyomo.solvers.plugins.solvers.mosek_direct
import pyomo.solvers.plugins.solvers.mosek_persistent
import pyomo.solvers.plugins.solvers.xpress_direct
import pyomo.solvers.plugins.solvers.xpress_persistent
